# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Rattail -> Rattail "versions" data import
"""

from __future__ import unicode_literals, absolute_import

import sqlalchemy_continuum as continuum

from rattail.db import model
from rattail.db.continuum import versioning_manager
from rattail.importing.rattail import FromRattailHandler, ToRattailHandler, FromRattail
from rattail.importing.model import ToRattail
from rattail.util import OrderedDict


class FromRattailToRattailVersions(FromRattailHandler, ToRattailHandler):
    """
    Handler for Rattail -> Rattail "versions" data import
    """
    local_title = "Rattail (Versions)"

    def get_importers(self):
        importers = OrderedDict()
        importers['Person'] = PersonImporter
        importers['PersonPhoneNumber'] = PersonPhoneNumberImporter
        importers['PersonEmailAddress'] = PersonEmailAddressImporter
        importers['PersonMailingAddress'] = PersonMailingAddressImporter
        importers['PersonNote'] = PersonNoteImporter
        importers['Role'] = RoleImporter
        # importers['Permission'] = PermissionImporter
        importers['User'] = UserImporter
        importers['UserRole'] = UserRoleImporter
        importers['Store'] = StoreImporter
        importers['StorePhoneNumber'] = StorePhoneNumberImporter
        importers['StoreEmailAddress'] = StoreEmailAddressImporter
        importers['Customer'] = CustomerImporter
        importers['CustomerPhoneNumber'] = CustomerPhoneNumberImporter
        importers['CustomerEmailAddress'] = CustomerEmailAddressImporter
        importers['CustomerMailingAddress'] = CustomerMailingAddressImporter
        importers['CustomerPerson'] = CustomerPersonImporter
        importers['CustomerGroup'] = CustomerGroupImporter
        importers['CustomerGroupAssignment'] = CustomerGroupAssignmentImporter
        importers['Member'] = MemberImporter
        importers['MemberPhoneNumber'] = MemberPhoneNumberImporter
        importers['MemberEmailAddress'] = MemberEmailAddressImporter
        importers['MemberMailingAddress'] = MemberMailingAddressImporter
        importers['Employee'] = EmployeeImporter
        importers['EmployeePhoneNumber'] = EmployeePhoneNumberImporter
        importers['EmployeeEmailAddress'] = EmployeeEmailAddressImporter
        importers['EmployeeStore'] = EmployeeStoreImporter
        importers['EmployeeDepartment'] = EmployeeDepartmentImporter
        importers['EmployeeHistory'] = EmployeeHistoryImporter
        importers['WorkedShift'] = WorkedShiftImporter
        importers['Vendor'] = VendorImporter
        importers['VendorPhoneNumber'] = VendorPhoneNumberImporter
        importers['VendorEmailAddress'] = VendorEmailAddressImporter
        importers['VendorContact'] = VendorContactImporter
        importers['Department'] = DepartmentImporter
        importers['Subdepartment'] = SubdepartmentImporter
        importers['Category'] = CategoryImporter
        importers['Family'] = FamilyImporter
        importers['ReportCode'] = ReportCodeImporter
        importers['DepositLink'] = DepositLinkImporter
        importers['Brand'] = BrandImporter
        importers['Tax'] = TaxImporter
        importers['Product'] = ProductImporter
        importers['ProductCode'] = ProductCodeImporter
        importers['ProductCost'] = ProductCostImporter
        importers['ProductPrice'] = ProductPriceImporter
        importers['LabelProfile'] = LabelProfileImporter
        return importers

    def begin_local_transaction(self):
        super(FromRattailToRattailVersions, self).begin_local_transaction()
        self.uow = versioning_manager.unit_of_work(self.session)
        self.transaction = self.uow.create_transaction(self.session)
        if self.comment:
            self.transaction.meta = {'comment': self.comment}

    def get_importer_kwargs(self, key, **kwargs):
        kwargs = super(FromRattailToRattailVersions, self).get_importer_kwargs(key, **kwargs)
        kwargs['transaction'] = self.transaction
        return kwargs


class VersionImporter(FromRattail, ToRattail):
    """
    Base class for version model importers
    """

    def get_model_class(self):
        return continuum.version_class(self.host_model_class)

    @property
    def simple_fields(self):
        fields = super(VersionImporter, self).simple_fields
        fields.remove('operation_type')
        fields.remove('transaction_id')
        fields.remove('end_transaction_id')
        return fields

    def cache_query(self):
        return self.session.query(self.model_class)\
                           .filter(self.model_class.end_transaction_id == None)\
                           .filter(self.model_class.operation_type != continuum.Operation.DELETE)

    def normalize_local_object(self, obj):
        data = super(VersionImporter, self).normalize_local_object(obj)
        # cache local (version) object, for updating end_transaction_id
        # (note that this is called for host side as well, hence check)
        if isinstance(obj, self.model_class):
            data['_object'] = obj
        return data

    def make_version(self, host_data, operation_type):
        key = self.get_key(host_data)
        with self.session.no_autoflush:
            version = self.new_object(key)
            self.populate(version, host_data)
            version.transaction = self.transaction
            version.operation_type = operation_type
            self.session.add(version)
            return version

    def populate(self, obj, data):
        for field in self.simple_fields:
            if field not in self.key and field in data and field in self.fields:
                setattr(obj, field, data[field])

    def create_object(self, key, host_data):
        return self.make_version(host_data, continuum.Operation.INSERT)

    def update_object(self, obj, host_data, local_data=None, **kwargs):
        last_version = local_data.pop('_object')
        last_version.end_transaction_id = self.transaction.id
        return self.make_version(host_data, continuum.Operation.UPDATE)
    

class PersonImporter(VersionImporter):
    host_model_class = model.Person

class PersonPhoneNumberImporter(VersionImporter):
    host_model_class = model.PersonPhoneNumber

class PersonEmailAddressImporter(VersionImporter):
    host_model_class = model.PersonEmailAddress

class PersonMailingAddressImporter(VersionImporter):
    host_model_class = model.PersonMailingAddress

class PersonNoteImporter(VersionImporter):
    host_model_class = model.PersonNote

class RoleImporter(VersionImporter):
    host_model_class = model.Role

# class PermissionImporter(VersionImporter):
#     host_model_class = model.Permission

class UserImporter(VersionImporter):
    host_model_class = model.User

class UserRoleImporter(VersionImporter):
    host_model_class = model.UserRole

class StoreImporter(VersionImporter):
    host_model_class = model.Store

class StorePhoneNumberImporter(VersionImporter):
    host_model_class = model.StorePhoneNumber

class StoreEmailAddressImporter(VersionImporter):
    host_model_class = model.StoreEmailAddress

class CustomerImporter(VersionImporter):
    host_model_class = model.Customer

class CustomerPhoneNumberImporter(VersionImporter):
    host_model_class = model.CustomerPhoneNumber

class CustomerEmailAddressImporter(VersionImporter):
    host_model_class = model.CustomerEmailAddress

class CustomerMailingAddressImporter(VersionImporter):
    host_model_class = model.CustomerMailingAddress

class CustomerPersonImporter(VersionImporter):
    host_model_class = model.CustomerPerson

class CustomerGroupImporter(VersionImporter):
    host_model_class = model.CustomerGroup

class CustomerGroupAssignmentImporter(VersionImporter):
    host_model_class = model.CustomerGroupAssignment

class MemberImporter(VersionImporter):
    host_model_class = model.Member

class MemberPhoneNumberImporter(VersionImporter):
    host_model_class = model.MemberPhoneNumber

class MemberEmailAddressImporter(VersionImporter):
    host_model_class = model.MemberEmailAddress

class MemberMailingAddressImporter(VersionImporter):
    host_model_class = model.MemberMailingAddress

class EmployeeImporter(VersionImporter):
    host_model_class = model.Employee

class EmployeePhoneNumberImporter(VersionImporter):
    host_model_class = model.EmployeePhoneNumber

class EmployeeEmailAddressImporter(VersionImporter):
    host_model_class = model.EmployeeEmailAddress

class EmployeeStoreImporter(VersionImporter):
    host_model_class = model.EmployeeStore

class EmployeeDepartmentImporter(VersionImporter):
    host_model_class = model.EmployeeDepartment

class EmployeeHistoryImporter(VersionImporter):
    host_model_class = model.EmployeeHistory

class WorkedShiftImporter(VersionImporter):
    host_model_class = model.WorkedShift

class VendorImporter(VersionImporter):
    host_model_class = model.Vendor

class VendorPhoneNumberImporter(VersionImporter):
    host_model_class = model.VendorPhoneNumber

class VendorEmailAddressImporter(VersionImporter):
    host_model_class = model.VendorEmailAddress

class VendorContactImporter(VersionImporter):
    host_model_class = model.VendorContact

class DepartmentImporter(VersionImporter):
    host_model_class = model.Department

class SubdepartmentImporter(VersionImporter):
    host_model_class = model.Subdepartment

class CategoryImporter(VersionImporter):
    host_model_class = model.Category

class FamilyImporter(VersionImporter):
    host_model_class = model.Family

class ReportCodeImporter(VersionImporter):
    host_model_class = model.ReportCode

class DepositLinkImporter(VersionImporter):
    host_model_class = model.DepositLink

class BrandImporter(VersionImporter):
    host_model_class = model.Brand

class TaxImporter(VersionImporter):
    host_model_class = model.Tax

class ProductImporter(VersionImporter):
    host_model_class = model.Product

class ProductCodeImporter(VersionImporter):
    host_model_class = model.ProductCode

class ProductCostImporter(VersionImporter):
    host_model_class = model.ProductCost

class ProductPriceImporter(VersionImporter):
    host_model_class = model.ProductPrice

class LabelProfileImporter(VersionImporter):
    host_model_class = model.LabelProfile



# -*- coding: utf-8; -*-
# -*- coding: utf-8; -*-
"""
Views for ${model_title} batches
"""

from __future__ import unicode_literals, absolute_import

import six

from rattail.db import model

from webhelpers2.html import tags

from tailbone.views.batch import FileBatchMasterView

from .handler import ${model_name}Handler


class ${model_name}View(FileBatchMasterView):
    """
    Master view for ${model_title} batches.
    """
    model_class = model.${model_name}
    batch_handler_class = ${model_name}Handler
    model_row_class = model.${model_name}Row
    model_title = "${model_title} Batch"
    model_title_plural = "${model_title} Batches"
    route_prefix = 'batch.${model_name.lower()}s'
    url_prefix = '/batch/${table_name.replace('_', '-')}'

    form_fields = [
        'vendor',
        'filename',
        'effective',
        'created',
        'created_by',
        'executed',
        'executed_by',
    ]

    row_labels = {
        'upc': "UPC",
    }

    row_grid_columns = [
        'sequence',
        'upc',
        'description',
        'status_code',
    ]

    def get_instance_title(self, batch):
        return six.text_type(batch.vendor)

    def configure_grid(self, g):
        super(${model_name}View, self).configure_grid(g)

        # vendor
        g.set_joiner('vendor', lambda q: q.outerjoin(model.Vendor))
        g.set_filter('vendor', model.Vendor.name,
                     default_active=True, default_verb='contains')
        g.set_sorter('vendor', model.Vendor.name)

    def configure_form(self, f):
        super(${model_name}View, self).configure_form(f)

        # readonly fields
        f.set_readonly('vendor')
        f.set_readonly('effective')

        # vendor
        f.set_renderer('vendor', self.render_vendor)

    def render_vendor(self, batch, field):
        vendor = batch.vendor
        if not vendor:
            return ""
        text = "({}) {}".format(vendor.id, vendor.name)
        url = self.request.route_url('vendors.view', uuid=vendor.uuid)
        return tags.link_to(text, url)

    def row_grid_row_attrs(self, row, i):
        attrs = {}
        if row.status_code == row.STATUS_SOME_CONCERN:
            attrs['class_'] = 'notice'
        if row.status_code == row.STATUS_UTTER_CHAOS:
            attrs['class_'] = 'warning'
        return attrs


def includeme(config):

    # fix permission group title
    config.add_tailbone_permission_group('${model_name.lower()}s', "${model_title}s")

    ${model_name}View.defaults(config)



# -*- coding: utf-8; -*-
"""add customer.wholesale

Revision ID: 5ef33e961026
Revises: e1edeba632f8
Create Date: 2018-11-17 19:16:57.206577

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '5ef33e961026'
down_revision = u'e1edeba632f8'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # customer
    op.add_column('customer', sa.Column('wholesale', sa.Boolean(), nullable=True))
    op.add_column('customer_version', sa.Column('wholesale', sa.Boolean(), autoincrement=False, nullable=True))


def downgrade():

    # customer
    op.drop_column('customer_version', 'wholesale')
    op.drop_column('customer', 'wholesale')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2018 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Batch Stuff
"""

from __future__ import unicode_literals, absolute_import

from six.moves import configparser

import six
import lockfile

from rattail.config import get_user_file


def consume_batch_id(source='RATAIL'):
    """
    Returns the next available batch identifier for ``source``, incrementing
    the number to preserve uniqueness.
    """
    path = get_user_file('rattail.conf', createdir=True)
    with lockfile.LockFile(path):

        parser = configparser.SafeConfigParser()
        parser.read(path)
        option = 'next_batch_id.{0}'.format(source)

        batch_id = 1
        if parser.has_section('rattail.sil'):
            if parser.has_option('rattail.sil', option):
                batch_id = parser.get('rattail.sil', option)
                batch_id = int(batch_id) if batch_id.isdigit() else 1

        if not parser.has_section('rattail.sil'):
            parser.add_section('rattail.sil')
        parser.set('rattail.sil', option, six.text_type(batch_id + 1))

        with open(path, 'wt') as f:
            parser.write(f)

    return '{0:08d}'.format(batch_id)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Trainwreck commands
"""

from __future__ import unicode_literals, absolute_import

import datetime
import sys
import logging

import sqlalchemy as sa

from rattail import __version__
from rattail import commands
from rattail.time import localtime, date_range
from rattail.util import load_object


log = logging.getLogger(__name__)


def main(*args):
    """
    Primary entry point for Trainwreck command system
    """
    if args:
        args = list(args)
    else:
        args = sys.argv[1:]

    cmd = Command()
    cmd.run(*args)


class Command(commands.Command):
    """
    Primary command for Trainwreck
    """
    name = 'trainwreck'
    version = __version__
    description = "Trainwreck -- Transaction data warehouse"
    long_description = ""


class ImportToTrainwreck(commands.ImportSubcommand):
    """
    Generic base class for commands which import *to* a Trainwreck DB.
    """
    # subclass must set these!
    handler_key = None
    default_handler_spec = None

    def get_handler_factory(self, **kwargs):
        if self.config:
            spec = self.config.get('rattail.trainwreck.importing', '{}.handler'.format(self.handler_key),
                                   default=self.default_handler_spec)
        else:
            # just use default, for sake of cmd line help
            spec = self.default_handler_spec
        return load_object(spec)


class Prune(commands.Subcommand):
    """
    Prune some dates from a Trainwreck database
    """
    name = 'prune'
    description = __doc__.strip()

    def add_parser_args(self, parser):
        parser.add_argument('--after', type=commands.date_argument, metavar='DATE',
                            help="Date *after* which all data should be pruned.  If set, no data "
                            "will be pruned from this or earlier dates.  If not set, there will be "
                            "no lower boundary for the prune.")
        parser.add_argument('--before', type=commands.date_argument, metavar='DATE',
                            help="Date *before* which all data should be pruned.  If set, no data "
                            "will be pruned from this or later dates.  If not set, will assume a "
                            "default of \"today\".")

        parser.add_argument('--dbkey', metavar='KEY', default='default',
                            help="Config key for the Trainwreck database engine to be used, i.e. "
                            "from which data should be pruned.  This key must be defined in the "
                            "[trainwreck.db] section of your config file.  Defaults to 'default'.")

        parser.add_argument('--dry-run', action='store_true',
                            help="Go through the full motions and allow logging etc. to "
                            "occur, but rollback (abort) the transaction at the end.")

    def make_trainwreck_session(self):
        """
        Must return a Trainwreck database session, which corresponds to the
        engine represented by the ``self.dbkey`` value.
        """
        from rattail.trainwreck.db import Session

        return Session(bind=self.config.trainwreck_engines[self.dbkey])

    def get_trainwreck_model(self):
        """
        Must return a reference to the Python module which contains Trainwreck
        data models.
        """
        return self.config.get_trainwreck_model()

    def get_earliest_date(self, session):
        """
        Fetches the earliest date which exists in the given session's database.
        Note that this is "calculated" according to ``Transaction.end_time``.
        """
        # TODO: not sure what might happen if there are no transactions with end_time ?
        trainwreck = self.get_trainwreck_model()
        end_time = session.query(sa.func.min(trainwreck.Transaction.end_time)).scalar()
        return localtime(self.config, end_time, from_utc=True).date()

    def prune_date(self, session, date, start_time, end_time):
        """
        Your custom command must implement the logic for this method.  It is
        responsible for fully pruning data from the given session's database,
        for the given date.  The start time and end times given may be used for
        your queries; they are in localtime and are zone-aware.
        """
        raise NotImplementedError

    def run(self, args):
        self.dbkey = args.dbkey
        session = self.make_trainwreck_session()

        if args.after:
            start_date = args.after + datetime.timedelta(days=1)
        else:
            # default is earliest available date
            start_date = self.get_earliest_date(session)

        if args.before:
            end_date = args.before - datetime.timedelta(days=1)
        else:
            # default is yesterday
            end_date = localtime(self.config).date() - datetime.timedelta(days=1)

        for date in date_range(start_date, end_date):
            log.info("pruning from '{}' Trainwreck for {}".format(args.dbkey, date))

            start_time = datetime.datetime.combine(start_date, datetime.time(0))
            start_time = localtime(self.config, start_time)

            end_time = datetime.datetime.combine(end_date + datetime.timedelta(days=1), datetime.time(0))
            end_time = localtime(self.config, end_time)

            self.prune_date(session, date, start_time, end_time)

        if args.dry_run:
            session.rollback()
            log.info("dry run, so transaction was rolled back")
        else:
            session.commit()
            log.info("transaction was committed")
        session.close()



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data models for reporting
"""

from __future__ import unicode_literals, absolute_import

import six
import sqlalchemy as sa

from rattail.db.model import Base, ExportMixin
from rattail.db.core import filename_column
from rattail.db.types import JSONTextDict
from rattail.util import NOTSET


@six.python_2_unicode_compatible
class ReportOutput(ExportMixin, Base):
    """
    Represents a generated report(s)
    """
    __tablename__ = 'report_output'
    model_title = "Generated Report"

    report_name = sa.Column(sa.String(length=255), nullable=True, doc="""
    Proper name of the report.
    """)

    report_type = sa.Column(sa.String(length=255), nullable=True, doc="""
    Type key for the report.
    """)

    params = sa.Column(JSONTextDict(), nullable=True, doc="""
    Parameters used to generate the report, encoded as JSON.
    """)

    filename = filename_column()

    def filepath(self, config, filename=NOTSET, **kwargs):
        if filename is NOTSET:
            filename = self.filename
        return super(ReportOutput, self).filepath(config, filename=filename, **kwargs)

    def __str__(self):
        return "{} {}".format(self.id_str, self.report_name)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2020 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

from __future__ import unicode_literals, absolute_import

import sys
import os.path
from setuptools import setup, find_packages


here = os.path.abspath(os.path.dirname(__file__))
exec(open(os.path.join(here, 'rattail', '_version.py')).read())
README = open(os.path.join(here, 'README.rst')).read()


requires = [
    #
    # Version numbers within comments below have specific meanings.
    # Basically the 'low' value is a "soft low," and 'high' a "soft high."
    # In other words:
    #
    # If either a 'low' or 'high' value exists, the primary point to be
    # made about the value is that it represents the most current (stable)
    # version available for the package (assuming typical public access
    # methods) whenever this project was started and/or documented.
    # Therefore:
    #
    # If a 'low' version is present, you should know that attempts to use
    # versions of the package significantly older than the 'low' version
    # may not yield happy results.  (A "hard" high limit may or may not be
    # indicated by a true version requirement.)
    #
    # Similarly, if a 'high' version is present, and especially if this
    # project has laid dormant for a while, you may need to refactor a bit
    # when attempting to support a more recent version of the package.  (A
    # "hard" low limit should be indicated by a true version requirement
    # when a 'high' version is present.)
    #
    # In any case, developers and other users are encouraged to play
    # outside the lines with regard to these soft limits.  If bugs are
    # encountered then they should be filed as such.
    #
    # package                           # low                   high

    'bcrypt',                           # 3.1.4
    'humanize',                         # 0.5.1
    'lockfile',                         # 0.9.1
    'openpyxl',                         # 2.5.0
    'packaging',                        # 19.0
    'progress',                         # 1.3
    'six',                              # 1.10.0
    'texttable',                        # 1.2.1
    'xlrd',                             # 1.1.0

    # TODO: Remove this / make it optional / etc.
    'Mako',                             # 1.0.0

    # Hardcode ``pytz`` minimum since apparently it isn't (any longer?) enough
    # to simply require the library.
    'pytz>=2013b',                      #                       2013b 
]


if sys.platform != 'win32':
    requires += [
        #
        # package                       # low                   high
            
        'pyinotify',                    # 0.9.3
    ]


extras = {

    'backup': [
        #
        # package                       # low                   high

        'msgpack',                      # 0.6.1
        'borgbackup',                   # 1.1.10
        'llfuse==1.3.4',                # 1.3.4                 1.3.4
    ],

    'bouncer': [
        #
        # package                       # low                   high

        # until we support PY3K we must stick with older flufl.bounce
        'flufl.bounce<3.0',             # 2.3
    ],

    'db': [
        #
        # package                       # low                   high

        # Support for multiple migration bases was added in Alembic 0.7.
        'alembic>=0.7.0',               #                       0.7.4

        'SQLAlchemy',                   # 0.7.6
        'SQLAlchemy-Continuum',         # 1.1.5
    ],

    'docs': [
        #
        # package                       # low                   high

        'Sphinx',                       # 1.1.3
    ],

    'tests': [
        #
        # package                       # low                   high

        'coverage',                     # 3.6
        'mock',                         # 1.0.1
        'nose',                         # 1.3.0
    ],
}


# TODO: pretty sure some/all of this should be merged with above
if sys.platform == 'win32':
    extras['auth'] = [
        #
        # package                       # low                   high

        'py-bcrypt-w32',                # 0.2.2
        ]
else:
    extras['auth'] = [
        #
        # package                       # low                   high

        'passlib',                      # 1.7.1
    ]


setup(
    name = "rattail",
    version = __version__,
    author = "Lance Edgar",
    author_email = "lance@edbob.org",
    url = "https://rattailproject.org/",
    license = "GNU GPL v3",
    description = "Retail Software Framework",
    long_description = README,

    classifiers = [
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Environment :: Win32 (MS Windows)',
        'Environment :: X11 Applications',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Topic :: Office/Business',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],

    install_requires = requires,
    extras_require = extras,
    tests_require = ['rattail[db,tests]'],
    test_suite = 'nose.collector',

    packages = find_packages(),
    include_package_data = True,
    zip_safe = False,

    entry_points = """

[console_scripts]
rattail = rattail.commands:main
rattail-dev = rattail.commands.dev:main
trainwreck = rattail.trainwreck.commands:main

[gui_scripts]
rattailw = rattail.commands:main

[rattail.commands]
backup = rattail.commands.backup:Backup
bouncer = rattail.commands.core:EmailBouncer
clonedb = rattail.commands.core:CloneDatabase
datasync = rattail.commands.core:DataSync
date-organize = rattail.commands.core:DateOrganize
execute-batch = rattail.commands.batch:ExecuteBatch
export-csv = rattail.commands.importing:ExportCSV
export-rattail = rattail.commands.importing:ExportRattail
filemon = rattail.commands.core:FileMonitorCommand
import-csv = rattail.commands.importing:ImportCSV
import-rattail = rattail.commands.importing:ImportRattail
import-sample = rattail.commands.importing:ImportSampleData
import-versions = rattail.commands.importing:ImportVersions
make-appdir = rattail.commands.core:MakeAppDir
make-batch = rattail.commands.batch:MakeBatch
make-config = rattail.commands.core:MakeConfig
make-user = rattail.commands.core:MakeUser
make-uuid = rattail.commands.core:MakeUUID
populate-batch = rattail.commands.batch:PopulateBatch
problems = rattail.commands.problems:Problems
purge-batches = rattail.commands.batch:PurgeBatches
purge-versions = rattail.commands.importing:PurgeVersions
refresh-batch = rattail.commands.batch:RefreshBatch
run-n-mail = rattail.commands.core:RunAndMail
runsql = rattail.commands.core:RunSQL
update-costs = rattail.commands.products:UpdateCosts
upgrade = rattail.commands.core:Upgrade

[rattail_dev.commands]
new-batch = rattail.commands.dev:NewBatch

[rattail.config.extensions]
rattail.db = rattail.db:ConfigExtension
rattail.trainwreck = rattail.trainwreck.config:TrainwreckConfig

[rattail.sil.column_providers]
rattail = rattail.sil.columns:provide_columns

[rattail.vendors.catalogs.parsers]
rattail.contrib.dutchvalley = rattail.contrib.vendors.catalogs.dutchvalley:DutchValleyCatalogParser
rattail.contrib.kehe = rattail.contrib.vendors.catalogs.kehe:KeheCatalogParser
rattail.contrib.lotuslight = rattail.contrib.vendors.catalogs.lotuslight:LotusLightCatalogParser
rattail.contrib.unfi = rattail.contrib.vendors.catalogs.unfi:UNFICatalogParser
rattail.contrib.unfi.2 = rattail.contrib.vendors.catalogs.unfi:UNFICatalogParser2

[rattail.vendors.invoices.parsers]
rattail.contrib.alberts = rattail.contrib.vendors.invoices.alberts:AlbertsInvoiceParser
rattail.contrib.kehe = rattail.contrib.vendors.invoices.kehe:KeheInvoiceParser
rattail.contrib.unfi = rattail.contrib.vendors.invoices.unfi:UnfiInvoiceParser

""",
)



# -*- coding: utf-8; -*-
"""add more versions

Revision ID: 3a500dc55e0f
Revises: 65098d24972f
Create Date: 2017-07-05 13:27:45.314766

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '3a500dc55e0f'
down_revision = u'65098d24972f'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # address_version
    op.create_table('address_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('parent_type', sa.String(length=20), autoincrement=False, nullable=True),
                    sa.Column('parent_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('preference', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('type', sa.String(length=15), autoincrement=False, nullable=True),
                    sa.Column('street', sa.String(length=100), autoincrement=False, nullable=True),
                    sa.Column('street2', sa.String(length=100), autoincrement=False, nullable=True),
                    sa.Column('city', sa.String(length=60), autoincrement=False, nullable=True),
                    sa.Column('state', sa.String(length=2), autoincrement=False, nullable=True),
                    sa.Column('zipcode', sa.String(length=10), autoincrement=False, nullable=True),
                    sa.Column('invalid', sa.Boolean(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_address_version_end_transaction_id'), 'address_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_address_version_operation_type'), 'address_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_address_version_transaction_id'), 'address_version', ['transaction_id'], unique=False)

    # email_version
    op.create_table('email_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('parent_type', sa.String(length=20), autoincrement=False, nullable=True),
                    sa.Column('parent_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('preference', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('type', sa.String(length=15), autoincrement=False, nullable=True),
                    sa.Column('address', sa.String(length=255), autoincrement=False, nullable=True),
                    sa.Column('invalid', sa.Boolean(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_email_version_end_transaction_id'), 'email_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_email_version_operation_type'), 'email_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_email_version_transaction_id'), 'email_version', ['transaction_id'], unique=False)

    # phone_version
    op.create_table('phone_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('parent_type', sa.String(length=20), autoincrement=False, nullable=True),
                    sa.Column('parent_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('preference', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('type', sa.String(length=15), autoincrement=False, nullable=True),
                    sa.Column('number', sa.String(length=20), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_phone_version_end_transaction_id'), 'phone_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_phone_version_operation_type'), 'phone_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_phone_version_transaction_id'), 'phone_version', ['transaction_id'], unique=False)

    # store_version
    op.create_table('store_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('id', sa.String(length=10), autoincrement=False, nullable=True),
                    sa.Column('name', sa.String(length=100), autoincrement=False, nullable=True),
                    sa.Column('database_key', sa.String(length=30), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_store_version_end_transaction_id'), 'store_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_store_version_operation_type'), 'store_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_store_version_transaction_id'), 'store_version', ['transaction_id'], unique=False)

    # customer_version
    op.create_table('customer_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('id', sa.String(length=20), autoincrement=False, nullable=True),
                    sa.Column('number', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('name', sa.String(length=255), autoincrement=False, nullable=True),
                    sa.Column('email_preference', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('active_in_pos', sa.Boolean(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_customer_version_end_transaction_id'), 'customer_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_customer_version_operation_type'), 'customer_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_customer_version_transaction_id'), 'customer_version', ['transaction_id'], unique=False)

    # customer_x_person_version
    op.create_table('customer_x_person_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('customer_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('person_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('ordinal', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_customer_x_person_version_end_transaction_id'), 'customer_x_person_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_customer_x_person_version_operation_type'), 'customer_x_person_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_customer_x_person_version_transaction_id'), 'customer_x_person_version', ['transaction_id'], unique=False)

    # customer_group_version
    op.create_table('customer_group_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('id', sa.String(length=20), autoincrement=False, nullable=True),
                    sa.Column('name', sa.String(length=255), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_customer_group_version_end_transaction_id'), 'customer_group_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_customer_group_version_operation_type'), 'customer_group_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_customer_group_version_transaction_id'), 'customer_group_version', ['transaction_id'], unique=False)

    # customer_x_group_version
    op.create_table('customer_x_group_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('customer_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('group_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('ordinal', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_customer_x_group_version_end_transaction_id'), 'customer_x_group_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_customer_x_group_version_operation_type'), 'customer_x_group_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_customer_x_group_version_transaction_id'), 'customer_x_group_version', ['transaction_id'], unique=False)

    # employee_version
    op.create_table('employee_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('id', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('person_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('status', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('display_name', sa.String(length=100), autoincrement=False, nullable=True),
                    sa.Column('full_time', sa.Boolean(), autoincrement=False, nullable=True),
                    sa.Column('full_time_start', sa.Date(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_employee_version_end_transaction_id'), 'employee_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_employee_version_operation_type'), 'employee_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_employee_version_transaction_id'), 'employee_version', ['transaction_id'], unique=False)

    # vendor_version
    op.create_table('vendor_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('id', sa.String(length=15), autoincrement=False, nullable=True),
                    sa.Column('name', sa.String(length=50), autoincrement=False, nullable=True),
                    sa.Column('special_discount', sa.Numeric(precision=5, scale=3), autoincrement=False, nullable=True),
                    sa.Column('lead_time_days', sa.Numeric(precision=5, scale=1), autoincrement=False, nullable=True),
                    sa.Column('order_interval_days', sa.Numeric(precision=5, scale=1), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_vendor_version_end_transaction_id'), 'vendor_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_vendor_version_operation_type'), 'vendor_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_vendor_version_transaction_id'), 'vendor_version', ['transaction_id'], unique=False)

    # vendor_contact_version
    op.create_table('vendor_contact_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('vendor_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('person_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('preference', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_vendor_contact_version_end_transaction_id'), 'vendor_contact_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_vendor_contact_version_operation_type'), 'vendor_contact_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_vendor_contact_version_transaction_id'), 'vendor_contact_version', ['transaction_id'], unique=False)


def downgrade():

    # vendor_contact_version
    op.drop_index(op.f('ix_vendor_contact_version_transaction_id'), table_name='vendor_contact_version')
    op.drop_index(op.f('ix_vendor_contact_version_operation_type'), table_name='vendor_contact_version')
    op.drop_index(op.f('ix_vendor_contact_version_end_transaction_id'), table_name='vendor_contact_version')
    op.drop_table('vendor_contact_version')

    # vendor_version
    op.drop_index(op.f('ix_vendor_version_transaction_id'), table_name='vendor_version')
    op.drop_index(op.f('ix_vendor_version_operation_type'), table_name='vendor_version')
    op.drop_index(op.f('ix_vendor_version_end_transaction_id'), table_name='vendor_version')
    op.drop_table('vendor_version')

    # employee_version
    op.drop_index(op.f('ix_employee_version_transaction_id'), table_name='employee_version')
    op.drop_index(op.f('ix_employee_version_operation_type'), table_name='employee_version')
    op.drop_index(op.f('ix_employee_version_end_transaction_id'), table_name='employee_version')
    op.drop_table('employee_version')

    # customer_x_group_version
    op.drop_index(op.f('ix_customer_x_group_version_transaction_id'), table_name='customer_x_group_version')
    op.drop_index(op.f('ix_customer_x_group_version_operation_type'), table_name='customer_x_group_version')
    op.drop_index(op.f('ix_customer_x_group_version_end_transaction_id'), table_name='customer_x_group_version')
    op.drop_table('customer_x_group_version')

    # customer_group_version
    op.drop_index(op.f('ix_customer_group_version_transaction_id'), table_name='customer_group_version')
    op.drop_index(op.f('ix_customer_group_version_operation_type'), table_name='customer_group_version')
    op.drop_index(op.f('ix_customer_group_version_end_transaction_id'), table_name='customer_group_version')
    op.drop_table('customer_group_version')

    # customer_x_person_version
    op.drop_index(op.f('ix_customer_x_person_version_transaction_id'), table_name='customer_x_person_version')
    op.drop_index(op.f('ix_customer_x_person_version_operation_type'), table_name='customer_x_person_version')
    op.drop_index(op.f('ix_customer_x_person_version_end_transaction_id'), table_name='customer_x_person_version')
    op.drop_table('customer_x_person_version')

    # customer_version
    op.drop_index(op.f('ix_customer_version_transaction_id'), table_name='customer_version')
    op.drop_index(op.f('ix_customer_version_operation_type'), table_name='customer_version')
    op.drop_index(op.f('ix_customer_version_end_transaction_id'), table_name='customer_version')
    op.drop_table('customer_version')

    # store_version
    op.drop_index(op.f('ix_store_version_transaction_id'), table_name='store_version')
    op.drop_index(op.f('ix_store_version_operation_type'), table_name='store_version')
    op.drop_index(op.f('ix_store_version_end_transaction_id'), table_name='store_version')
    op.drop_table('store_version')

    # phone_version
    op.drop_index(op.f('ix_phone_version_transaction_id'), table_name='phone_version')
    op.drop_index(op.f('ix_phone_version_operation_type'), table_name='phone_version')
    op.drop_index(op.f('ix_phone_version_end_transaction_id'), table_name='phone_version')
    op.drop_table('phone_version')

    # email_version
    op.drop_index(op.f('ix_email_version_transaction_id'), table_name='email_version')
    op.drop_index(op.f('ix_email_version_operation_type'), table_name='email_version')
    op.drop_index(op.f('ix_email_version_end_transaction_id'), table_name='email_version')
    op.drop_table('email_version')

    # address_version
    op.drop_index(op.f('ix_address_version_transaction_id'), table_name='address_version')
    op.drop_index(op.f('ix_address_version_operation_type'), table_name='address_version')
    op.drop_index(op.f('ix_address_version_end_transaction_id'), table_name='address_version')
    op.drop_table('address_version')



# -*- coding: utf-8 -*-
"""add inventory_reason.hidden

Revision ID: 765ce0ae5bbd
Revises: a33583135905
Create Date: 2018-06-01 12:52:46.968644

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '765ce0ae5bbd'
down_revision = u'a33583135905'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # invadjust_reason
    op.add_column('invadjust_reason', sa.Column('hidden', sa.Boolean(), nullable=True))


def downgrade():

    # invadjust_reason
    op.drop_column('invadjust_reason', 'hidden')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
DataSync for Linux
"""

from __future__ import unicode_literals, absolute_import

import sys
import time
import datetime
import logging
from traceback import format_exception

from rattail.db import api
from rattail.daemon import Daemon
from rattail.threads import Thread
from rattail.datasync.config import load_profiles
from rattail.datasync.util import get_lastrun, get_lastrun_setting, get_lastrun_timefmt, next_batch_id
from rattail.mail import send_email
from rattail.time import make_utc, localtime


log = logging.getLogger(__name__)


class DataSyncDaemon(Daemon):
    """
    Linux daemon implementation of DataSync.
    """

    def run(self):
        """
        Starts watcher and consumer threads according to configuration.
        """
        for key, profile in load_profiles(self.config).items():

            # Create watcher thread for the profile.
            name = '{}-watcher'.format(key)
            log.debug("starting thread '{}' with watcher: {}".format(name, profile.watcher_spec))
            thread = Thread(target=watch_for_changes, name=name, args=(self.config, profile.watcher))
            thread.daemon = True
            thread.start()

            # Create consumer threads, unless watcher consumes itself.
            if not profile.watcher.consumes_self:

                # Create a thread for each "isolated" consumer.
                # for consumer in profile.isolated_consumers:
                for consumer in profile.consumers:
                    name = '{}-consumer-{}'.format(key, consumer.key)
                    log.debug("starting thread '%s' with consumer: %s", name, consumer.spec)
                    thread = Thread(target=consume_changes_perpetual, name=name, args=(self.config, profile, consumer))
                    thread.daemon = True
                    thread.start()

        # Loop indefinitely.  Since this is the main thread, the app will
        # terminate when this method ends; all other threads are "subservient"
        # to this one.
        while True:
            time.sleep(.01)


def watch_for_changes(config, watcher):
    """
    Target for DataSync watcher threads.
    """
    # Let watcher do any setup it needs.
    watcher.setup()

    # the 'last run' value is maintained as zone-aware UTC
    lastrun = get_lastrun(config, watcher.key)
    lastrun_setting = get_lastrun_setting(config, watcher.key)
    timefmt = get_lastrun_timefmt(config)

    while True:

        thisrun = make_utc(tzinfo=True)
        attempts = 0
        errtype = None
        while True:

            attempts += 1

            try:
                changes = watcher.get_changes(lastrun)

            except Exception as error:

                exc_type, exc, traceback = sys.exc_info()

                # If we've reached our final attempt, stop retrying.
                if attempts >= watcher.retry_attempts:
                    log.warning("attempt #{} failed calling `watcher.get_changes()`, this thread "
                                "will now *terminate* until datasync restart".format(attempts),
                                exc_info=True)
                    send_email(config, 'datasync_error_watcher_get_changes', {
                        'watcher': watcher,
                        'error': exc,
                        'attempts': attempts,
                        'traceback': ''.join(format_exception(exc_type, exc, traceback)).strip(),
                        'datasync_url': config.datasync_url(),
                    })
                    raise

                # If this exception is not the first, and is of a different type
                # than seen previously, do *not* continue to retry.
                if errtype is not None and not isinstance(error, errtype):
                    log.exception("new exception differs from previous one(s), "
                                  "giving up on watcher.get_changes()")
                    raise

                # Record the type of exception seen, and pause for next retry.
                errtype = type(error)
                log.warning("attempt #{} failed for '{}' watcher.get_changes()".format(attempts, watcher.key))
                log.debug("pausing for {} seconds before making attempt #{} of {}".format(
                    watcher.retry_delay, attempts + 1, watcher.retry_attempts))
                if watcher.retry_delay:
                    time.sleep(watcher.retry_delay)

            else:
                # watcher got changes okay, so record/process and prune, then
                # break out of inner loop to reset the attempt count for next grab
                lastrun = thisrun
                api.save_setting(None, lastrun_setting, lastrun.strftime(timefmt))
                if changes:
                    log.debug("got {} changes from '{}' watcher".format(len(changes), watcher.key))
                    # TODO: and what if this step fails..??
                    if record_or_process_changes(watcher, changes):
                        prune_changes(watcher, changes)
                break

        # sleep a moment between successful change grabs
        time.sleep(watcher.delay)


def record_or_process_changes(watcher, changes):
    """
    Now that we have changes from the watcher, we'll either record them as
    proper DataSync changes, or else let the watcher process them (if it
    consumes self).
    """
    from rattail.db import Session, model

    # If watcher consumes itself, then it will process its own changes.  Note
    # that there are no assumptions made about the format or structure of these
    # change objects.
    if watcher.consumes_self:
        session = Session()
        try:
            watcher.process_changes(session, changes)
        except Exception:
            log.exception("error calling watcher.process_changes()")
            session.rollback()
            raise
        else:
            session.commit()
            log.debug("watcher has consumed its own changes")
            return True
        finally:
            session.close()

    else:
        # Give all change stubs the same timestamp, to help identify them
        # as a "batch" of sorts, so consumers can process them as such.
        # (note, this is less important for identifiying a batch now that we
        # have batch_id, but is probably still helpful anyway)
        now = datetime.datetime.utcnow()

        # Save change stub records to rattail database, for consumer thread
        # to find and process.
        saved = 0
        session = Session()
        # assign new/unique batch_id so that consumers can keep things straight
        batch_id = next_batch_id(session)
        batch_seq = 0
        for key, change in changes:
            batch_seq += 1
            for consumer in watcher.consumer_stub_keys:
                session.add(model.DataSyncChange(
                    source=watcher.key,
                    batch_id=batch_id,
                    batch_sequence=batch_seq,
                    payload_type=change.payload_type,
                    payload_key=change.payload_key,
                    deletion=change.deletion,
                    obtained=now,
                    consumer=consumer))
                saved += 1
            session.flush()
        session.commit()
        session.close()
        log.debug("saved {} '{}' changes to datasync queue".format(saved, watcher.key))
        return True


def prune_changes(watcher, changes):
    """
    Tell the watcher to prune the original change records from its source
    database, if relevant.
    """
    if watcher.prunes_changes:
        try:
            # Note that we only give it the keys for this.
            pruned = watcher.prune_changes([c[0] for c in changes])
        except:
            log.exception("error calling watcher.prune_changes()")
            raise
        else:
            if pruned is not None:
                log.debug("pruned {} changes from '{}' database".format(pruned, watcher.key))


def consume_changes_perpetual(config, profile, consumer):
    """
    Target for DataSync consumer thread.
    """
    # tell consumer to do initial setup
    consumer.setup()

    # begin thread perma-loop
    while True:

        # try to consume all current changes
        if not consume_current_changes(config, profile, consumer):

            # consumption failed, so exit the perma-loop
            # (this thread is now dead)
            break

        # wait 1 sec by default, then look for more changes
        time.sleep(consumer.delay)


def consume_current_changes(config, profile, consumer):
    """
    Consume all changes currently available.
    """
    from rattail.db import Session

    session = Session()
    model = config.get_model()

    # determine first 'obtained' timestamp
    first = session.query(model.DataSyncChange)\
                   .filter(model.DataSyncChange.source == consumer.watcher.key)\
                   .filter(model.DataSyncChange.consumer == consumer.key)\
                   .order_by(model.DataSyncChange.obtained)\
                   .first()

    error = False
    while first:

        # try to consume these changes
        if not consume_changes_from(config, session, consumer, first.obtained):
            error = True
            break

        else: # fetch next 'obtained' timestamp
            first = session.query(model.DataSyncChange)\
                           .filter(model.DataSyncChange.source == profile.key)\
                           .filter(model.DataSyncChange.consumer == consumer.key)\
                           .order_by(model.DataSyncChange.obtained)\
                           .first()

    # no more changes! (or perhaps an error)
    session.close()
    return not error


def consume_changes_from(config, session, consumer, obtained):
    """
    Consume all changes which were "obtained" at the given timestamp.

    Note that while this function must be given a Rattail database session, the
    function logic is responsible for committing (or rolling back) the session
    transaction.  The assumption here is that this session will be used *only*
    for managing the datasync change queue, and that we should be "saving our
    progress" as changes are fully consumed, i.e. those changes are deleted
    from the queue via session commit.
    """
    model = config.get_model()

    # we only want changes "obtained" at the given time.  however, at least
    # until all code has been refactored, we must take two possibilities into
    # account here: some changes may have been given a batch ID, but some may
    # not.  we will prefer those with batch ID, or fall back to those without.
    changes = session.query(model.DataSyncChange)\
                     .filter(model.DataSyncChange.source == consumer.watcher.key)\
                     .filter(model.DataSyncChange.consumer == consumer.key)\
                     .filter(model.DataSyncChange.obtained == obtained)\
                     .filter(model.DataSyncChange.batch_id != None)\
                     .order_by(model.DataSyncChange.batch_id,
                               model.DataSyncChange.batch_sequence)\
                     .all()
    if changes:
        # okay, we got some with a batch ID, now we must prune that list down
        # so that we're only dealing with a single batch
        batch_id = changes[0].batch_id
        changes = [c for c in changes if c.batch_id == batch_id]
    else:
        # no changes with batch ID, so let's get all without ID instead
        changes = session.query(model.DataSyncChange)\
                         .filter(model.DataSyncChange.source == consumer.watcher.key)\
                         .filter(model.DataSyncChange.consumer == consumer.key)\
                         .filter(model.DataSyncChange.obtained == obtained)\
                         .filter(model.DataSyncChange.batch_id == None)\
                         .all()

    # maybe limit size of batch to process.  this can be useful e.g. when large
    # amounts of changes land in the queue with same timestamp, and versioning
    # is also enabled.
    batch_size = config.getint('rattail.datasync', 'batch_size_limit')
    if batch_size and len(changes) > batch_size:
        changes = changes[:batch_size]

    log.debug("will process %s changes from %s", len(changes),
              localtime(config, obtained, from_utc=True))
    consumer.begin_transaction()

    attempts = 0
    errtype = None
    while True:

        attempts += 1

        try:
            consumer.process_changes(session, changes)
            session.flush()

        except Exception as errobj: # processing failed!
            exc_type, exc, traceback = sys.exc_info()

            consumer.rollback_transaction()
            session.rollback()

            # if we've reached our final attempt, stop retrying
            if attempts >= consumer.retry_attempts:
                log.warning("attempt #%s failed calling `consumer.process_changes()`; "
                            "this thread will now *terminate* until datasync restart",
                            attempts, exc_info=True)
                send_email(config, 'datasync_error_consumer_process_changes', {
                    'watcher': consumer.watcher,
                    'consumer': consumer,
                    'error': exc,
                    'attempts': attempts,
                    'traceback': ''.join(format_exception(exc_type, exc, traceback)).strip(),
                    'datasync_url': config.datasync_url(),
                })
                return False

            # if this exception is not the first, and is of a different type
            # than seen previously, do *not* continue to retry
            if errtype is not None and not isinstance(errobj, errtype):
                log.exception("new exception differs from previous one(s), "
                              "giving up on consumer.process_changes()")
                return False

            # record the type of exception seen; maybe pause before next retry
            errtype = type(errobj)
            log.warning("attempt #%s failed for '%s' -> '%s' consumer.process_changes()",
                        attempts, consumer.watcher.key, consumer.key)
            log.debug("pausing for %s seconds before making attempt #%s of %s",
                      consumer.retry_delay, attempts + 1, consumer.retry_attempts)
            if consumer.retry_delay:
                time.sleep(consumer.retry_delay)

        else: # consumer processed changes okay

            consumer.commit_transaction()

            # delete these changes from datasync queue
            for i, change in enumerate(changes):
                session.delete(change)
                if i % 200 == 0:
                    session.flush()
            session.flush()
            session.commit()

            # can stop the attempt/retry loop now
            log.debug("processed %s changes", len(changes))
            break

    return True



# -*- coding: utf-8; -*-
"""add product.default_pack

Revision ID: 51756b0edcdf
Revises: a295dee8d1e9
Create Date: 2018-07-03 19:56:26.827825

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '51756b0edcdf'
down_revision = u'a295dee8d1e9'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # product
    op.add_column('product', sa.Column('default_pack', sa.Boolean(), nullable=True))
    op.add_column('product_version', sa.Column('default_pack', sa.Boolean(), autoincrement=False, nullable=True))


def downgrade():

    # product
    op.drop_column('product_version', 'default_pack')
    op.drop_column('product', 'default_pack')



# -*- coding: utf-8 -*-
"""add product.case_pack

Revision ID: 268ff6454410
Revises: 571c070b7080
Create Date: 2015-02-11 19:59:50.344482

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '268ff6454410'
down_revision = u'571c070b7080'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():
    op.add_column('product', sa.Column('case_pack', sa.Integer(), nullable=True))
    op.add_column('product_version', sa.Column('case_pack', sa.Integer(), nullable=True))


def downgrade():
    op.drop_column('product_version', 'case_pack')
    op.drop_column('product', 'case_pack')
